"""
    aiotinydb - asyncio compatibility shim for tinydb

    Copyright 2017 Pavel Pletenev <cpp.create@gmail.com>
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


class AIOTinyDBError(Exception):
    'Base class for AIOTinyDB exceptions'
    pass


class DatabaseNotReady(AIOTinyDBError):
    'Indicates wrong usage of AIOTinyDB'
    pass


class ReadonlyStorageError(AIOTinyDBError):
    'Raised when a readonly storage is attempted to be written to'
    pass


class NotOverridableError(AIOTinyDBError):
    'Indicates a non-overridable method'
    pass



"""
    aiotinydb - asyncio compatibility shim for tinydb

    Copyright 2017 Pavel Pletenev <cpp.create@gmail.com>
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
# pylint: disable=super-init-not-called
import os
from abc import abstractmethod
from tinydb.storages import Storage, JSONStorage
from aiofilelock import AIOMutableFileLock, AIOImmutableFileLock

from .exceptions import NotOverridableError, ReadonlyStorageError


class AIOStorage(Storage):
    """
    Abstract asyncio Storage class
    """
    @abstractmethod
    async def __aenter__(self):
        """
        Initialize storage in async manner (open files, connections, etc...)
        """
        raise NotImplementedError('To be overridden!')

    @abstractmethod
    async def __aexit__(self, exc_type, exc, traceback):
        """
        Finalize storage in async manner (close files, connections, etc...)
        """
        raise NotImplementedError('To be overridden!')

    def close(self):
        """
        This is not called and should NOT be used
        """
        raise NotOverridableError('NOT to be overridden or called, use __aexit__!')


class AIOJSONStorage(AIOStorage, JSONStorage):
    """
    Asyncronous JSON Storage for AIOTinyDB
    """
    def __init__(self, filename, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self._filename = filename
        self._lock = None
        self._handle = None

    async def __aenter__(self):
        if self._handle is None:
            try:
                self._handle = open(self._filename, 'r+')
            except FileNotFoundError:
                dirname = os.path.dirname(self._filename)
                if dirname:
                    os.makedirs(dirname, exist_ok=True)
                self._handle = open(self._filename, 'w+')
            self._lock = AIOMutableFileLock(self._handle)
            await self._lock.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        if self._handle is not None:
            await self._lock.__aexit__(exc_type, exc, traceback)
            self._lock = None
            self._handle.close()


class AIOImmutableJSONStorage(AIOJSONStorage):
    """
    Asyncronous readonly JSON Storage for AIOTinyDB
    """
    async def __aenter__(self):
        if self._handle is None:
            self._handle = open(self._filename, 'r')
            self._lock = AIOImmutableFileLock(self._handle)
            await self._lock.__aenter__()
        return self

    def write(self, data):
        raise ReadonlyStorageError('AIOImmutableJSONStorage cannot be written to')



"""
    aiotinydb - asyncio compatibility shim for tinydb

    Copyright 2017 Pavel Pletenev <cpp.create@gmail.com>
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# pylint: disable=super-init-not-called,arguments-differ
from tinydb import TinyDB
from .exceptions import NotOverridableError, DatabaseNotReady
from .storage import AIOJSONStorage


class AIOTinyDB(TinyDB):
    """
    TinyDB for asyncio

    # Example
    ```
    import asyncio
    from aiotinydb import AIOTinyDB

    async def test():
        async with AIOTinyDB('test.json') as db:
            db.insert(dict(counter=1))

    loop = asyncio.new_event_loop()
    loop.run_until_complete(test())
    loop.close()
    ```
    """
    DEFAULT_STORAGE = AIOJSONStorage
    def __init__(self, *args, **kwargs):
        self._storage_cls = kwargs.pop('storage', self.DEFAULT_STORAGE)
        self._table_name = kwargs.pop('default_table', self.DEFAULT_TABLE)
        self._args = args
        self._kwargs = kwargs
        self._table_cache = {}
        self._storage = None
        self._table = None

    def purge_table(self, name):
        if self._storage is None:
            raise DatabaseNotReady('File is not opened. Use with AIOTinyDB(...):')
        return super().purge_table(name)

    def purge_tables(self):
        if self._storage is None:
            raise DatabaseNotReady('File is not opened. Use with AIOTinyDB(...):')
        return super().purge_tables()

    def table(self, name='_default', **options):
        if self._storage is None:
            raise DatabaseNotReady('File is not opened. Use with AIOTinyDB(...):')
        return super().table(name='_default', **options)

    def tables(self):
        if self._storage is None:
            raise DatabaseNotReady('File is not opened. Use with AIOTinyDB(...):')
        return super().tables()

    def __getattr__(self, val):
        if self._storage is None:
            raise AttributeError('File is not opened. Use with AIOTinyDB(...):')
        return super().__getattr__(val)

    async def __aenter__(self):
        if self._storage is None:
            self._storage = self._storage_cls(*self._args, **self._kwargs)
            await self._storage.__aenter__()
            self._table = self.table(self._table_name)
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        if self._storage:
            await self._storage.__aexit__(exc_type, exc, traceback)
            self._storage = None
            self._table_cache = {}

    def close(self):
        raise NotOverridableError('Usual methods will not work on async')

    def __enter__(self):
        raise NotOverridableError('Usual methods will not work on async')

    def __exit__(self, exc_type, exc, tb):
        raise NotOverridableError('Usual methods will not work on async')



"""
    aiotinydb - asyncio compatibility shim for tinydb

    Copyright 2017 Pavel Pletenev <cpp.create@gmail.com>
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from setuptools import setup

setup(
    name='aiotinydb',
    version='1.0.0',
    author='Pavel Pletenev',
    author_email='cpp.create@gmail.com',
    url='https://github.com/ASMfreaK/aiotinydb',
    license='GPLv3',
    description='asyncio compatibility shim for tinydb',
    packages=['aiotinydb'],
    install_requires=['aiofilelock', 'tinydb'],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
    ],
)
